@@ -1,4 +1,3 @@
-
 """
 boson.py
 ====================================
@@ -99,6 +98,10 @@ class Boson(Core):
 
                 if self.conn.is_open:
                     self.logger.info("Connected")
+        else:
+            self.connect(port, baudrate)
+            if self.conn.is_open:
+                self.logger.info("Connected")
 
     @classmethod
     def find_serial_device(self):
@@ -140,7 +143,7 @@ class Boson(Core):
         if sys.platform.startswith('win32'):
             device_check_path = pkg_resources.resource_filename('flirpy', 'bin/find_cameras.exe')
             device_id = int(subprocess.check_output([device_check_path, "FLIR Video"]).decode())
-
+            self.logger.info("Device ID: {}".format(device_id))
             if device_id >= 0:
                 return device_id
 
@@ -176,13 +179,15 @@ class Boson(Core):
                     pass
             
             # For some reason multiple devices can show up
+            # The lower index is often correct, so do a sort first
             if len(dev) > 1:
-                for d in dev:
+                for d in sorted(dev):
                     cam = cv2.VideoCapture(d + cv2.CAP_V4L2)
                     data = cam.read()
                     cam.release()
 
-                    if data is not None:
+                    # data[1] is the image and should be a numpy array
+                    if data[0] == True and data[1] is not None:
                         res = d
                         break
             elif len(dev) == 1:
@@ -223,9 +228,11 @@ class Boson(Core):
         self.cap.set(cv2.CAP_PROP_FRAME_HEIGHT,256)
         
         # The order of these calls matters!
+        # The Y16 command tells the Boson to send pre-AGC 16 bit pixels. This means the pixel intensity is linearly
+        # proportional to the flux incident on the pixel. This also means all internal image processing done by the
+        # boson is ignored. There are other options, see Boson datasheet.
         self.cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc(*"Y16 "))
-        self.cap.set(cv2.CAP_PROP_CONVERT_RGB, False)
-        
+        self.cap.set(cv2.CAP_PROP_CONVERT_RGB, 0)
 
     def grab(self, device_id=None):
         """
@@ -256,6 +263,10 @@ class Boson(Core):
         return image
 
     def release(self):
+        """
+        This will release the camera hardware.
+        :return:
+        """
         if self.cap:
             self.cap.release()
 
@@ -268,7 +279,6 @@ class Boson(Core):
 
         return 
 
-
     def get_sensor_serial(self):
         """
         Get the serial number of the sensor
@@ -283,7 +293,10 @@ class Boson(Core):
         res = self._send_packet(function_id, receive_size=4)
         res = self._decode_packet(res, receive_size=4)
 
-        return struct.unpack(">I", res)[0]
+        if res is not None and len(res) == 4:
+            res = struct.unpack(">I", res)[0]
+
+        return res
     
     def get_firmware_revision(self):
         """
@@ -300,7 +313,10 @@ class Boson(Core):
         res = self._send_packet(function_id, receive_size=12)
         res = self._decode_packet(res, receive_size=12)
 
-        return struct.Struct(">iii").unpack_from(res)
+        if res is not None and len(res) == 12:
+             res = struct.Struct(">iii").unpack_from(res)
+
+        return res
     
     def get_part_number(self):
         """
@@ -315,7 +331,10 @@ class Boson(Core):
         res = self._send_packet(function_id, receive_size=20)
         res = self._decode_packet(res, receive_size=20)
 
-        return res.decode("utf-8") 
+        if res is not None and len(res) == 20:
+            res = res.decode("utf-8")
+
+        return res
 
     def do_ffc(self):
         """
@@ -352,17 +371,21 @@ class Boson(Core):
         res = self._send_packet(function_id, receive_size=2)
         res = self._decode_packet(res, receive_size=2)
 
-        return struct.unpack(">H", res)[0]
+        if res is not None and len(res) == 2:
+            res = struct.unpack(">H", res)[0]
+
+        return res
 
     def get_ffc_mode(self):
         """
         Get the current FFC mode
 
-        0 = FLR_BOSON_NO_FFC_PERFORMED
-        1 = FLR_BOSON_FFC_IMMINENT
-        2 = FLR_BOSON_FFC_IN_PROGRESS
-        3 = FLR_BOSON_FFC_COMPLETE
-        4 = FLR_BOSON_FFCSTATUS_END
+        FLR_BOSON_MANUAL_FFC = 0
+        FLR_BOSON_AUTO_FFC = 1
+        FLR_BOSON_EXTERNAL_FFC = 2
+        FLR_BOSON_SHUTTER_TEST_FFC = 3
+        FLR_BOSON_FFCMODE_END = 4
+
 
         Returns
         -------
@@ -374,8 +397,60 @@ class Boson(Core):
         res = self._send_packet(function_id, receive_size=4)
         res = self._decode_packet(res, receive_size=4)
 
+        if res is not None and len(res) == 4:
+            res = struct.unpack(">I", res)[0]
+
+        return res
+
+    def get_gao_ffc_mode(self):
+        """
+        Get whether the ffc correction is applied in the Boson image processing pipeline.
+        0 = disabled
+        1 = enabled
+
+
+        Returns
+        -------
+
+            int
+                FLR_ENABLE_E
+        """
+        function_id = 0x00000004
+        res = self._send_packet(function_id, receive_size=4)
+        res = self._decode_packet(res, receive_size=4)
+
         return struct.unpack(">I", res)[0]
 
+    def get_ffc_desired(self):
+        """
+        Get the state of the FFC desired flag. 0 = not desired. 1 = desired.
+        :return: int 0 or 1
+        """
+        function_id = 0x00050055
+        res = self._send_packet(function_id, receive_size=4)
+        res = self._decode_packet(res, receive_size=4)
+
+        return struct.unpack(">I", res)[0]
+
+    def get_nuc_desired(self):
+        """
+        Get the state of the NUC table switch desired flag. 0 = not desired. 1 = desired.
+        :return: int 0 or 1
+        """
+        function_id = 0x0005005F
+        res = self._send_packet(function_id, receive_size=2)
+        res = self._decode_packet(res, receive_size=2)
+
+        return struct.unpack(">H", res)[0]
+
+    def do_nuc_table_switch(self):
+        """
+        This will perform a table NUC table switch, if the nuc table switch desired flag is set.
+        :return:
+        """
+        function_id = 0x00050050
+        self._send_packet(function_id)
+
     def set_ffc_auto(self):
         """
         Set the FFC mode to automatic
@@ -428,7 +503,10 @@ class Boson(Core):
         res = self._send_packet(function_id, receive_size=2)
         res = self._decode_packet(res, receive_size=2)
 
-        return struct.unpack(">H", res)[0]/10.0
+        if res is not None and len(res) == 2:
+            res = struct.unpack(">H", res)[0]/10.0
+
+        return res
     
     def set_ffc_frame_threshold(self, seconds):
         """
@@ -454,7 +532,10 @@ class Boson(Core):
         res = self._send_packet(function_id, receive_size=4)
         res = self._decode_packet(res, receive_size=4)
 
-        return struct.unpack(">I", res)[0]
+        if res is not None and len(res) == 4:
+            res = struct.unpack(">I", res)[0]
+
+        return res
     
     def get_last_ffc_temperature(self):
         """
@@ -465,8 +546,10 @@ class Boson(Core):
         res = self._send_packet(function_id, receive_size=2)
         res = self._decode_packet(res, receive_size=2)
 
-        return struct.unpack(">H", res)[0]/10.0
+        if res is not None and len(res) ==2:
+            res = struct.unpack(">H", res)[0]/10.0
 
+        return res
 
     def get_last_ffc_frame_count(self):
         """
@@ -478,8 +561,39 @@ class Boson(Core):
         res = self._send_packet(function_id, receive_size=4)
         res = self._decode_packet(res, receive_size=4)
 
-        return struct.unpack(">I", res)[0]
-    
+        if res is not None and len(res) == 4:
+            res = struct.unpack(">I", res)[0]
+
+        return res
+
+    def set_num_ffc_frame(self, num_frame):
+        """
+        Set the number of frames for the camera to integrate during FFC. 8 is factory default. With averager on, the
+        time to perform ffc is doubled.
+        :param num_frame: int 2, 4, 8, 16.
+        :return:
+        """
+        function_id = 0x0000000D
+        command = struct.pack(">H", num_frame)
+        res = self._send_packet(function_id, data=command)
+        res = self._decode_packet(res)
+
+        return
+
+    def get_num_ffc_frame(self):
+        """
+        Get the number of frames the camera integrates during FFC.
+        :return: int: 2,4,6, or 8
+        """
+        function_id = 0x0000000E
+        res = self._send_packet(function_id, receive_size=2)
+        res = self._decode_packet(res)
+        
+        if res is not None and len(res) == 2:
+            res = struct.unpack(">H",res)[0]
+            
+        return res
+
     def get_frame_count(self):
         """
         Get the number of frames captured since the camera was turned on.
@@ -494,7 +608,10 @@ class Boson(Core):
         res = self._send_packet(function_id, receive_size=4)
         res = self._decode_packet(res, receive_size=4)
 
-        return struct.unpack(">I", res)[0]
+        if res is not None and len(res) == 4:
+            res = struct.unpack(">I", res)[0]
+
+        return res
 
     def get_fpa_temperature(self):
         """
@@ -510,11 +627,14 @@ class Boson(Core):
         res = self._send_packet(function_id, receive_size=2)
         res = self._decode_packet(res, receive_size=2)
 
-        return struct.unpack(">H", res)[0]/10.0
+        if res is not None and len(res) == 2:
+            res = struct.unpack(">H", res)[0]/10.0
+
+        return res
 
     def get_camera_serial(self):
         """
-        Get the camera serial number 
+        Get the camera serial number
 
         Returns
         -------
@@ -526,8 +646,64 @@ class Boson(Core):
         res = self._send_packet(function_id, receive_size=4)
         res = self._decode_packet(res, receive_size=4)
 
-        return struct.unpack(">I", res)[0]
-    
+        if res is not None and len(res) == 4:
+            res = struct.unpack(">I", res)[0]
+
+        return res
+
+    def set_pwr_on_defaults(self):
+        """
+        Apply the current settings as power on defaults.
+        :return:
+        """
+        function_id = 0x00050018
+        self._send_packet(function_id)
+
+        return
+
+    def set_pwr_on_defaults_factory(self):
+        """
+        This will restore power on defaults to factory settings. The camera should be disconnected and reconnected to
+        make sure all changes take effect. A reboot does not appear to be enough.
+        :return:
+        """
+        function_id = 0x0005001B
+        self._send_packet(function_id)  # Read factory boot settings and set them as current settings
+        self.set_pwr_on_defaults()  # Set the factory boot settings as power on settings.
+        self.logger.warn("consider cycling power and reconnecting to camera to insure all factory settings take effect.")
+
+        return
+
+    def set_averager(self, value):
+        """
+        This will set the smart averager state to the value
+        :param value: int 0=off (factory default), 1=on
+        :return:
+        """
+        function_id = 0x0000000B
+        command = struct.pack(">I", value)
+        # Tell the camera to set the smart averager state
+        self._send_packet(function_id, data=command)
+        # But this change does not happen immediately. Instead it must be saved as a power on default.
+        self.set_pwr_on_defaults() # Set the current states as power on defaults
+        self.logger.warn("The camera must be disconnected and reconnected for the change in averager state to take effect.")
+
+        return
+
+    def get_averager(self):
+        """
+        Get the current state of the smart averager.
+        :return: int. 0=off 1=on.
+        """
+        function_id = 0x0000000C
+        res = self._send_packet(function_id, receive_size=4)
+        res = self._decode_packet(res, receive_size=4)
+        
+        if res is not None and len(res) == 4:
+            res = struct.unpack(">I", res)[0]
+
+        return res
+
     def _decode_packet(self, data, receive_size=0):
         """
         Decodes a data packet from the camera.
@@ -547,21 +723,31 @@ class Boson(Core):
 
         """
         payload = None
-        payload_len = len(data) - 17
+        payload_len = len(data) - 15
 
         if payload_len > 0:
-            frame = struct.Struct(">BBIII{}sHB".format(payload_len))
-            res = frame.unpack(data)
-
-            start_marker, channel_id, sequence, function_id, return_code, payload, crc, end_marker = res
+            try:
+                frame = struct.Struct(">BBIII{}sB".format(payload_len))
+                res = frame.unpack(data)
+                start_marker, channel_id, sequence, function_id, return_code, payload, end_marker = res
+            except Exception as e:
+                self.logger.error(str(e))
+                self.logger.error("Failed to unpack payload")
+                return None
         elif payload_len < 0:
-            raise ValueError
+            self.logger.error("Failed to open payload")
+            return None
         else:
-            frame = struct.Struct(">BBIIIHB")
-            res = frame.unpack(data)
+            try:
+                frame = struct.Struct(">BBIIIHB")
+                res = frame.unpack(data)
+
+                start_marker, channel_id, sequence, function_id, return_code, crc, end_marker = res
+            except Exception as e:
+                self.logger.error(str(e))
+                self.logger.error("Failed to unpack payload")
+                return None
 
-            start_marker, channel_id, sequence, function_id, return_code, crc, end_marker = res
-        
         if return_code == 0x0203:
             self.logger.warning("Boson response: range error")
         elif return_code == 0x017F:
@@ -590,12 +776,18 @@ class Boson(Core):
         unstuffed_payload = None
 
         if payload_len > 0:
-            unstuffed_payload = self._unstuff(payload)
+            unstuffed_payload = self._unstuff(payload) # data + crc
+
+            frame = struct.Struct(">{}sH".format(len(unstuffed_payload)-2)) # split data, crc
+            unstuffed_payload, crc = frame.unpack(unstuffed_payload)
 
         crc_bytes = self._crc(header_bytes, unstuffed_payload)
 
         if crc != crc_bytes:
-            self.logger.warning("Invalid checksum")
+            self.logger.debug("Got crc {}, but expected {}".format(crc, crc_bytes))
+            self.logger.debug("Payload: {}".format(payload))
+            self.logger.debug("Frame: {}".format(data))
+            self.logger.warning("Invalid checksum, but data may be OK")
 
         return unstuffed_payload
 
@@ -606,7 +798,7 @@ class Boson(Core):
         temp = bytearray()
 
         for byte in data:
-            
+
             if sys.version_info[0] < 3:
                 byte = ord(byte)
 
@@ -631,7 +823,7 @@ class Boson(Core):
         temp = bytearray()
         unstuff = False
 
-        for byte in data:
+        for i, byte in enumerate(data):
 
             if sys.version_info[0] < 3:
                 byte = ord(byte)
@@ -640,7 +832,12 @@ class Boson(Core):
                 temp.append(byte + 0xD)
                 unstuff = False
             elif byte == 0x9E:
-                unstuff = True
+                if i == (len(data) - 1):
+                    continue
+                elif data[i+1] in [0x81, 0x91, 0xA1]:
+                    unstuff = True
+                else:
+                    continue
             else:
                 temp.append(byte)
 
@@ -661,7 +858,7 @@ class Boson(Core):
 
             if time.time() - tstart > timeout:
                 break
-        
+
         return frame
 
     def _crc(self, header, payload=None):
@@ -672,7 +869,7 @@ class Boson(Core):
 
         if payload is not None:
             data += bytes(payload)
-        
+
         return binascii.crc_hqx(data, 0x1D0F)
 
     def _send_packet(self, function_id, data=None, receive_size=0):
@@ -714,10 +911,12 @@ class Boson(Core):
 
         if data is not None:
             payload += self._bitstuff(data)
-        
+
         payload += footer_bytes
 
         self.send(payload)
         res = self.receive()
 
-        return res
\ No newline at end of file
+        return res
+
+
