@@ -28,12 +28,19 @@ class Core:
         if self.conn is not None:
             if self.conn.is_open:
                 self.conn.close()
-    
+
+    def release(self):
+        pass
+        
     def close(self):
         self.disconnect()
+        self.release()
 
     def __enter__(self):
         return self
 
     def __exit__(self, exc_type, exc_value, traceback):
+        self.close()
+
+    def __del__(self):
         self.close()
\ No newline at end of file
